% function_FLL_OptimizationLinearThreshold.m
%
% V. Katkovnik - Tampere University of Technology - July 2006
% -----------------------------------------------------------------------
%
% Performs the simplified linear optimization of the thresholds dzeta-k for the FLL adaptive scale 
% selection algorithmt
% It is done for the exponential distributions: Gaussian, Poissonian, Binary Birnoulli

% Demo for paper: Vladimir Katkovnik,  Vladimir Spokoiny "SPATIALLY ADAPTIVE NON-GAUSSIAN IMAGING VIA FITTED LOCAL
% LIKELIHOOD TECHNIQUE", Proceedings of SMMSP 2006, Florence.

% z : noisy observation
% y : true image (assumed as unknown)

%
% Other key variables:
% y_hat   : anisotropic "fused" estimate
% y_hat_Q : adaptive directional estimate
% h_opt_Q : adaptive directional scales
%
%

clear all
close all


directional_resolution=1;  % number of directions
fusing=2;                  % fusing type   (1 inverse-variance weighted fusing, 2 piecewise regular)
%%% Type of distribution %%%%%%%%%%%%%%%%%%%%
%%% 1 indicate the selected type of distribution 
gaussian=0; %% pdf=1
poisson=1;  %% pdf=2
binary=0;   %% pdf=3
kappa=1;  % default parameter latter changed for specific distribution
filtering=1; % control of the filtering in FLL adaptive scale selection
% filtering=1 means filtering on; filtering=0 means that the filtering is  off
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% alpha and r are the parameters of the procedure
% default:
r=1/2; alpha=1; % for Bernoulli
r=1; alpha=1; % good for Poissonian and gaussian

% Grid for dzeta_1 selection

H_dzetaq_1=[.1:.1:3]; %


filtering=1;

tic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%--------------------------------------------------------------------------
% DISTRIBUTIONS PARAMETERS and KERNELS SCALES
%--------------------------------------------------------------------------
if poisson==1
    pdf=2; 
    kappa=25.5;
  
h1=floor((1.5).^[1:7]);

h2=max(1,ceil(h1*.75*tan(1*pi/8)));  % CAMERAMANrow vectors h1 and h2 need to have the same lenght
%h2=ones(size(h1));
end
if gaussian==1
   
kappa=1;

% h1=[1 2 3 5 9 19];
h1=[1 2 3 5 7 11];
h1=floor((1.5).^[1:7]);
% h2=ones(size(h1));
h2=max(1,ceil(h1*.75*tan(1*pi/8)));  % row vectors h1
pdf=1;
sigma_noise=0.1
end

if binary==1
   
kappa=1;
delta=.9; % parameter of the model

%h1=[1 2 3 5 9 19];
h1=floor((1.5).^[1:7]);
%h2=ones(size(h1));
h2=max(1,ceil(h1*tan(.75*pi/8)));  % row vectors h1 and h2 need to have the same lenght

pdf=3;

end

lenh=length(h1);
K=length(h1);
%--------------------------------------------------------------------------
% WINDOWS PARAMETERS
%--------------------------------------------------------------------------
sig_winds=[.5*ones(size(h1)); .5*ones(size(h2))];    % Gaussian parameter
beta=1;                     % Parameter of window 6

window_type=112;  % window=1 for uniform, window=2 for Gaussian
% window=6 for exponentions with beta
% window=8 for Interpolation
% window=11 for sectorial windows

window=112; % for sectorial uniform windows
 % h2=max(1,ceil(h1*tan(0.5*pi/directional_resolution))); % for sectorial windows


TYPE=10;            % TYPE IS A SYMMETRY OF THE WINDOW
% 00 SYMMETRIC
% 10 NONSYMMETRIC ON X1 and SYMMETRIC ON X2
% 11 NONSYMMETRIC ON X1,X2  (Quadrants)
%
% for rotated directional kernels the method that is used for rotation can be specified by adding
% a binary digit in front of these types, as follows:
%
% 10
% 11  ARE "STANDARD" USING NN (Nearest Neighb.) (you can think of these numbers with a 0 in front)
% 00
%
% 110
% 111  ARE EXACT SAMPLING OF THE EXACT ROTATED KERNEL
% 100
%
% 210
% 211  ARE WITH BILINEAR INTERP
% 200
%
% 310
% 311  ARE WITH BICUBIC INTERP (not reccomended)
% 300


%--------------------------------------------------------------------------
% MODELLING OF THE INVARIANT SIGNAL 
%--------------------------------------------------------------------------
yy=im2double(imread('image_Cameraman256.png'));
y=ones(size(yy)/2);

tic
%---------------------------------------------------------
% Images SIMULATION
%---------------------------------------------------------
[size_z_1,size_z_2]=size(y);

state = 2055615866;
rand('state', state);
randn('state', state);

if gaussian==1
n=repmat(sigma_noise,size_z_1,size_z_2).*randn(size(y));
z = y + n;
else
z = y;
end

if poisson==1
z=poissrnd(kappa*y)/kappa;
end

if binary==1
 
y=y/max(y(:));
kappa=1;
pdf=3;
z1=unifrnd(zeros(size(y)),ones(size(y)));
y=0.5*(1-delta)+y*delta;
II=(z1<y);
z=zeros(size(y));
z(II==1)=1;
%    
%     y=(y-.5)*delta+0.5;
%     z = binornd(1,y);
end   
sigma=function_stdEst2D(z);   %%% estimates noise standard deviation
disp('Initial (noisy) observation criteria values:')
function_Errors(y,z,z,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%---------------------------------------------------------
% Kernels construction
%---------------------------------------------------------

% calling kernel creation function
[kernels, kernels_higher_order]=function_CreateLPAKernels([0 0],h1,h2,TYPE,window_type,1,sig_winds,beta);
disp('  ')
disp('kernels created');

sigmaiter=repmat(sigma,size_z_1,size_z_2);
stop_condition=0;
toc


clear yh h_opt_Q y_hat_Q var_opt_Q stdh

YICI_Final1=0; var_inv=0;         YICI_Final2=0;
CWW=0;
CWW2=0;


%htbar=timebar('LPA-ICI running','Progress');
%%%%% loops %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  YICI_Final1=0;  YICI_Final2=0;

for s1=1:1  % DIRECTIONS
           
    for s2=1:lenh     % kernel size index
        gh=kernels_higher_order{s1,s2,1}(:,:,1);   %gets single kernel from the cell array
        ghorigin(s1,s2)=gh((end+1)/2,(end+1)/2);
        bound1=min([(find(sum(gh~=0,2)));abs(find(sum(gh~=0,2))-size(gh,1)-1)]); % removes unnecessary zeroes
        bound2=min([(find(sum(gh~=0,1))),abs(find(sum(gh~=0,1))-size(gh,2)-1)]); % removes unnecessary zeroes
        gh=gh(bound1:size(gh,1)-bound1+1,bound2:size(gh,2)-bound2+1);            % removes unnecessary zeroes
        yh(1:size_z_1,1:size_z_2,s2)= conv2(z,gh,'same'); % Estimation
        %yh(1:size_z_1,1:size_z_2,s2)= conv2(z+10000,gh,'same')-10000; % Estimation
        N_Kulback(s2)=sum(gh(:)>0);
        N_Kulback(s2)=h1(s2);
        
       stdh(:,:,s2)=repmat(sigma*(sum(gh(:).^2))^0.5,size_z_1,size_z_2);  % Std of the estimate
    end %% for s2, window sizes
end
%
    
   %%%%%%%%%%%% FIND Ideal right hand side value %%%%%%%%%%%%%%%    
       [N1,N2]=size(y); edge_x=max(h1)+1;
       for s_lenh=1:lenh
        yh1=yh(:,:,s_lenh); %stdh1=stdh(:,:,1); 
        gammaICI=ones(size(h1));
        dzeta1_array_ideal=zeros(size(y));
        dzeta1_array_ideal_0=function_Kullback(yh1*kappa,y*kappa,sigma,pdf)*N_Kulback(s_lenh);
        dzeta1_array_ideal(1:N1-2*edge_x+1,1:N2-2*edge_x+1)=dzeta1_array_ideal_0(edge_x:N1-edge_x,edge_x:N2-edge_x);

    Cr_h1(s_lenh)=sum(sum( abs(dzeta1_array_ideal).^r))/(N1-2*edge_x+1)/(N2-2*edge_x+1);
       end % END s_lenh
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
 %%%%%%%%%%%%%%% FIND dzeta 1 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 
         gammaICI=[1 ones(1,lenh-1)*10000];
        for SS_K=2:2
         clear FLL_dzeta11    FLL_dzeta1
             
         dzeta1_array_estimate0=zeros(size(y));
       for s_dzeta1=[1:length(H_dzetaq_1)]
           
      gammaICI(SS_K)=H_dzetaq_1(s_dzeta1);
      if filtering==1
  [YICI,h_opt,std_opt]=function_FLL_for_ML_WOS(yh,stdh,gammaICI(2:end),N_Kulback,sigma,pdf,kappa,2*(s1-1)*pi/directional_resolution);
  
      else
  [YICI,h_opt,std_opt]=function_FLL_for_ML_WOS(yh,stdh,gammaICI(2:end),N_Kulback,sigma,pdf,kappa);
   
      end
  
  for s_k=SS_K:lenh
          for s_l=SS_K:s_k
  dzeta1_array_estimate0=...
  (h_opt(:,:)==s_l-1).*function_Kullback(yh(:,:,s_k)*kappa,yh(:,:,s_l-1)*kappa,sigma,pdf)*N_Kulback(s_k);
  dzeta1_array_estimate(1:N1-2*edge_x+1,1:N2-2*edge_x+1)=dzeta1_array_estimate0(edge_x:N1-edge_x,edge_x:N2-edge_x);
  FLL_dzeta_sl(s_dzeta1,s_l)=sum(sum(abs(dzeta1_array_estimate).^r))/(N1-2*edge_x+1)/(N2-2*edge_x+1);
          end % END s_l
     FLL_dzeta1(s_dzeta1,s_k-SS_K+1)=sum(FLL_dzeta_sl(s_dzeta1,SS_K:end));
   end % END s_k
       end % END s_dzeta1
       
       for sd=1:length(H_dzetaq_1)
           FLL_dzeta11(sd)=max(FLL_dzeta1(sd,:));
        end
     %  FLL_dzeta11=max(FLL_dzeta1');
       Cr_max=max(Cr_h1);
       figure(1), plot(H_dzetaq_1,FLL_dzeta11), hold on , plot(H_dzetaq_1,ones(size(H_dzetaq_1))* Cr_max*alpha/(K-1))
 
   % keyboard
   tt1=(FLL_dzeta11<=ones(size(H_dzetaq_1))*Cr_max*alpha/(K-1));
     if tt1==0
          gammaICI(SS_K)=max(H_dzetaq_1);
          'tt=0'
     else
         [kk1]=find(tt1'==1);
      gammaICI(SS_K)=H_dzetaq_1(min(kk1));
     end
   
          end %% SS_K

    
   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  ' dzeta1_value', dzeta1_value=gammaICI(2: SS_K)
   
%H_gamma_min=[1 :-.02:.02 ]*dzeta1_value; % Grid for slope selection
H_gamma_min=[1. :-.02:.02 ]*dzeta1_value; % Grid for slope selection

     for s_slope=[1:length(H_gamma_min)]
     delta_gamma=(dzeta1_value-H_gamma_min(s_slope))/(K-2);
     gammaICI(2:K)=dzeta1_value*ones(1,K-1)-delta_gamma*([1:K-1]-1);
     if filtering==1
     [YICI,h_opt,std_opt]=function_FLL_for_ML_WOS(yh,stdh,gammaICI(2:end),N_Kulback,sigma,pdf,kappa,2*(s1-1)*pi/directional_resolution);
     else
          [YICI,h_opt,std_opt]=function_FLL_for_ML_WOS(yh,stdh,gammaICI(2:end),N_Kulback,sigma,pdf,kappa);
     end
     ssigma=zeros(N1,N2);
     for s_k=2:K
     ssigma0=(function_Kullback(yh(:,:,s_k)*kappa,YICI(:,:)*kappa,sigma,pdf)*N_Kulback(s_k)).^r;
  
     ssigma(1:N1-2*edge_x+1,1:N2-2*edge_x+1)=ssigma0(edge_x:N1-edge_x,edge_x:N2-edge_x);
     SIGMA_K(s_k,s_slope)=sum(sum(real(ssigma)))/(N1-2*edge_x+1)/(N2-2*edge_x+1); 
        
   end % s_k

   end % s_slope

   %%%%%%%%% NOVEL PART %%%%%%%%%
  for s_slope=1:length(H_gamma_min)
          SIGMA_K_max(s_slope)=max(SIGMA_K(:,s_slope)); % maximum on s_k
       end
     
   tt=(SIGMA_K_max<=ones(size(H_gamma_min))*Cr_max*alpha);
     if sum(tt)==0
          gamma_min=min(H_gamma_min);
          'tt=0'
     end
     if sum(tt)==length(tt)
         gamma_min=max(H_gamma_min);
     end
         if ((sum(tt)==0)|(sum(tt)==length(tt)))==0
     
         [kk1]=find(tt'==1);
     
      gamma_min=H_gamma_min(max(kk1));
         end
     
      delta_gamma=(dzeta1_value-gamma_min)/(K-2);
     gammaICI(2:K)=dzeta1_value*ones(1,K-1)-delta_gamma*([1:K-1]-1);
    
   'gammaICI(2:end)',  gammaICI(2:end)
    figure(3), plot(H_gamma_min,SIGMA_K_max), hold on , plot(H_gamma_min,ones(size(H_gamma_min))* Cr_max*alpha)

    toc